module Main where
--Some examples of using Inductive Programming (IP) on
--Anderson's (Fisher's) Iris data.
--Compile thus: ghc --make Iris

-- Here we have:
--   Unsupervised clustering (mixture modelling) under various within-cluster
--     distributions. NB best overall is 2 clusters with the
--     'custom' model which allows for correlated variables;
--     it even beats the "true" mixture model.
--   Also modelling of the 3 "known" species
--     (a) as the "true" mixture-model (for various distributions), and
--     (b) as a segmentation model, adding in log(n-1)+log(n-2)-1 bits for
--        the "known" cut-points (by rights should search for them), and a
--   Classification-tree to predict the species from the 4 measurements
--     (it's all down to petal width).
--
-- Note that a mixture model must state the class-memberships but
-- these are implicit by position (and cheaper) in a segmentation.
-- A clustering is independent of the order of the training data which
-- is definitely not the case for a segmentation.

import IP

type Flower = (SepalLength, SepalWidth, PetalLength, PetalWidth, Species)

type SepalLength = Double
type SepalWidth  = Double
type PetalLength = Double
type PetalWidth  = Double

dataAcc = 0.1   --cts data accuracy, +/-0.05cm, i.e., to the nearest mm.

data Species = IS | IVe | IVi deriving (Bounded, Enum, Eq, Show)


instance Read Species where                                                --IO
  readsPrec n str =
   [check str2Species ["Iris-setosa", "Iris-versicolor", "Iris-virginica"] str]

startsWith xs ys =
--Hopes to find ys at the beginning of xs.
  let (ys', xs') = splitAt (length ys) xs
  in if ys == ys' then Just xs' else Nothing

check f (w:ws) xs =
--Check that xs starts with one of a list of alternatives;
--if so, return f(the match) and the rest of xs.
  case xs `startsWith` w of
    Just xs' -> (f w, xs')
    Nothing  -> check f ws xs
check _ _ _  = error "in check"

str2Species "Iris-setosa"     = IS
str2Species "Iris-versicolor" = IVe
str2Species "Iris-virginica"  = IVi

species2str IS  = "Iris-setosa"
species2str IVe = "Iris-versicolor"
species2str IVi = "Iris-virginica"

-- ----------------------------------------------------------------------------
estNorm = estNormalWeighted 0 10 0.1 5 dataAcc    --with reasonably bland prior

variate4 m0 m1 m2 m3 =
--Turn 4 Models into a Model of 4-tuples.
  let nlP (d0,d1,d2,d3) =
        nlPr m0 d0 + nlPr m1 d1 + nlPr m2 d2 + nlPr m3 d3
  in MnlPr (msg1 m0 + msg1 m1 + msg1 m2 + msg1 m3)
           nlP
           (\() -> "[v4 " ++ show [m0,m1,m2,m3] ++ "]")

estVariate4Weighted e0 e1 e2 e3 ds ws =
--Turn 4 estimators into an estimator of a Model of 4-tuples.
  let (d0s, d1s, d2s, d3s) = unzip4 ds
  in variate4 (e0 d0s ws) (e1 d1s ws) (e2 d2s ws) (e3 d3s ws)

estNx4 =                           --estimate 4 independent normal variates
  estVariate4Weighted estNorm estNorm estNorm estNorm


--The following is a simple way to allow for correlation twixt length & width.
estNx4' = estBivariateWeighted                 -- ((@0,@1|@0),(@2,@3|@2))
 ((estConditionWeighted                              -- (@0,@1|@0)
    estNorm                                            -- @0
    (estLinear1Weighted (-5) 5 0.1 5 dataAcc)),        -- @1|@0
  (estConditionWeighted                              -- (@2,@3|@2)
    estNorm                                            -- @2
    (estLinear1Weighted (-5) 5 0.1 5 dataAcc)))        -- @3|@2


custom4Model m0 fm1 fm2 fm3 =                          --       @0
  let nlP (d0,d1,d2,d3) =                              --      /  \
       nlPr m0 d0 + condNlPr fm1 d0 d1                 --    @1    @2
       + condNlPr fm2 d0 d2 + condNlPr fm3 d2 d3       --           \
  in MnlPr (msg1 m0 + msg1 fm1 + msg1 fm2 + msg1 fm3)  --            @3
     nlP
     (\() -> "[c4 " ++ show m0 ++ show [fm1,fm2,fm3] ++ "]")

estCustom4Weighted e0 e1 e2 e3 ds ws =
  let (d0s, d1s, d2s, d3s) = unzip4 ds
  in custom4Model (e0 d0s ws) (e1 d0s d1s ws) (e2 d0s d2s ws) (e3 d2s d3s ws)

estCustom4 = estCustom4Weighted
  estNorm                                    -- @0
  (estLinear1Weighted (-5) 5 0.1 5 dataAcc)  -- @1|@0
  (estLinear1Weighted (-5) 5 0.1 5 dataAcc)  -- @2|@0
  (estLinear1Weighted (-5) 5 0.1 5 dataAcc)  -- @3|@2

-- ----------------------------------------------------------------------------
cluster est dataSet n =                                   --try up to n classes
--Tries 1,...,n class (cluster) mixture-models for a given estimator.
 let clstr m =
       if m > n then putStrLn( "--\n")
       else
         let mx = estMixture (take m (repeat est)) dataSet
         in putStrLn( " " ++ show m ++ " classes: " ++ show mx )
         >> putStrLn( "  msgLen = " ++ show(msgBase 2 mx dataSet )
                   ++ " = " ++ show(msg1Base 2 mx)
                   ++ " + " ++ show(msg2Base 2 mx dataSet) ++ " bits" )
         >> clstr (m+1)
 in putStrLn "clustering"
 >> clstr 1


known est dataSet =              -- we humans know that there are three species
--This creates the "true" 3-class mixture Model for a given estimator.
  let n  = fromIntegral(length dataSet)
      w1 = repeat 1
      ds = [d1, d2, d3]
      d1 = take 50          dataSet   --IS,  known Iris-setosa
      d2 = take 50 (drop 50 dataSet)  --IVe, known Iris-versicolor
      d3 = drop 100         dataSet   --IVi, known Iris-virginica
      ms = [m1, m2, m3]
      m1 = est d1 w1
      m2 = est d2 w1
      m3 = est d3 w1
      cs  = zipWith (msgBase 2) ms ds  --costs
      [c1, c2, c3] = cs
      mx = mixture (Mix (freqs2model [50,50,50]) [m1,m2,m3])
  in putStrLn "known (3 classes taken as read)"
  >> putStrLn( "  1 IS : " ++ show m1 ++ show(c1) ++ " bits")
  >> putStrLn( "  2 IVe: " ++ show m2 ++ show(c2) ++ " bits")
  >> putStrLn( "  3 IVi: " ++ show m3 ++ show(c3) ++ " bits")
  >> putStrLn( "  mixture msgLen = "
      ++ show(msgBase  2 mx dataSet) ++ " = "
      ++ show(msg1Base 2 mx) ++ " + "
      ++ show(msg2Base 2 mx dataSet) ++ " bits (cheat)")
  >> putStrLn( "  segmentation msgLen = "
      ++ show(logBase 2 (n-1) + logBase 2 (n-2) - 1 + sum cs) ++ " bits" )
      -- ignores the cost of stating 'nSegments = 2'
  >> putStrLn "--\n"

-- -----------------------------------LA--5/2008--http://www.allisons.org/ll/--
main = putStrLn topline
 >> putStrLn "-- Examples on Fisher's / Anderson's Iris data --\n"
 >> readFile "Data/iris.data"  -- Fisher's / Anderson's Iris data

 >>= \str ->
 let flowers = read str :: [Flower]
     (sepLen, sepWid, petLen, petWid, species) = unzip5 flowers
     dataSet4  = zip4 sepLen sepWid petLen petWid
     dataSet4' = zip (zip sepLen sepWid) (zip petLen petWid)
 in putStrLn "Independent normals:\n"
 >> known   estNx4  dataSet4      --uses background knowledge
 >> cluster estNx4  dataSet4  4                                    --clustering
 >> putStrLn "Lengths~widths, normals:\n"
 >> known   estNx4' dataSet4'
 >> cluster estNx4' dataSet4' 4
 >> putStrLn "Custom, petal~sepal, lengths~widths, normals:\n"
 >> known   estCustom4  dataSet4
 >> cluster estCustom4  dataSet4  4

 >>
 let ipOp = zip dataSet4 species
     ct = estCTree estMultiState splits dataSet4 species  --classification tree
 in putStrLn "Tree for (Species | sepal,petal x length,width):"
 >> putStrLn(" tree: " ++ show ct)
 >> putStrLn("  msgLen = "
     ++ show(msgBase 2 (functionModel2model ct) ipOp)  --NB. species only
     ++ " = " ++ show(msg1Base 2 ct) ++ "+"
     ++ show(msg2Base 2 (functionModel2model ct) ipOp)
     ++ ", v. null: " ++ show(150 * logBase 2 3) ++ " bits")

 >> putStrLn "\n-- done --"
-- ----------------------------------------------------------------------------
