-- See:
-- Lloyd ALLISON. Models for machine learning and data mining in
--   functional programming.       doi:10.1017/S0956796804005301
--   J. Functional Programming, 15(1), pp.15-32, Jan. 2005 (online 23 Jul 2004)
-- Lloyd ALLISON. Types and classes of machine learning and data mining.
--   ACSC2003, pp.207-215, February 2003.
-- Lloyd ALLISON. A programming paradigm for machine learning, with
--   a case study of Bayesian networks.  ACSC2006, January 2006.
--
-- Author: Lloyd ALLISON
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License (GPL) as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.  This program is distributed in the hope
-- that it will be useful, but without any warranty, without even the implied
-- warranty of merchantability or fitness for a particular purpose.  See the
-- GNU General Public License for more details.  You should have received a
-- copy of the GNU General Public License with this program; if not, write to:
-- Free Software Foundation, Inc., Boston, MA 02111, USA.

module IP.CTrees (module IP.CTrees) where

import IP.Utilities
import IP.Classes
import IP.Models


-- A classification tree, a.k.a. decision tree, is
-- an instance of classes SuperModel and FunctionModel.


data CTreeType inSpace opSpace =
  CTleaf   (ModelType opSpace)  |
  CTfork   MessageLength (Splitter inSpace) [CTreeType inSpace opSpace]  |
  CTforkFM (FunctionModelType inSpace Int)  [CTreeType inSpace opSpace]
  -- The last option, CTFork, is rather speculative.

  -- CTleaf   (mdl opSpace)  |  -- want  data (Model mdl) =>...  but that
  -- leads to Non-type variables in constraint: SuperModel (mdl opSpace),
  -- which is not H'98.


instance SuperModel (CTreeType inSpace opSpace) where
 -- NB. For simplicity, this costs the *structure* at 1-bit per node
 -- which is only optimal for binary trees. See Wallace & Patrick 1993.
 -- It also assumes that a tree uses only Trad or only non-Trad (FM) forks.
 msg1 (CTleaf leafModel)       = log2 + msg1 leafModel
 msg1 (CTfork fnLen f dts) = log2 + fnLen + (foldl (+) 0 (map msg1 dts))
 -- pity no built-in class Function in Haskell
 msg1 (CTforkFM fnMixer dts)     =
    log2 + (msg1 fnMixer)+(foldl (+) 0 (map msg1 dts))


instance FunctionModel CTreeType where
  condModel (CTleaf leafModel)       i = leafModel
  condModel (CTfork fnLen f dts) i = condModel (dts !! (applySplitter f i)) i

  -- only include the following to check the types
  condModel (CTforkFM fnMixer dts)     i =   -- below is == the 1-branch method
    let choice = biggest( map (\n -> condPr fnMixer i n)
                              [0 .. ((length dts)-1)]   ) 0 (-1) 0
        biggest  []    posn bigVal bigPosn = bigPosn
        biggest (p:ps) posn bigVal bigPosn =
          if p > bigVal then biggest ps (posn+1) p      posn
                        else biggest ps (posn+1) bigVal bigPosn
    in condModel (dts !! choice) i


instance Show (CTreeType inSpace opSpace) where                -- printing etc.
  showsPrec d (CTleaf mdl) rest = "{CTleaf " ++ (showsPrec d mdl ("}" ++ rest))
  showsPrec d (CTfork _ f subTrees) rest =
    "{CTfork " ++ (show f) ++ (showsPrec d subTrees ("}" ++ rest))
  showsPrec d (CTforkFM f subTrees) rest =
    "{CTforkFM " ++ (show f) ++ (showsPrec d subTrees ("}" ++ rest))


                                     -- estimate a classification tree, a CTree
estCTree  estLeafMdl splits  ipSet opSet =
 let  -- NB. estLeafMdl must be able to handle an empty data set
  search ipSet opSet =
   let
    leaf    = CTleaf leafMdl        -- the simplest 1-level tree is just a leaf
    leafMdl = estLeafMdl opSet
    leafMsg = msg (functionModel2model leaf) (zip ipSet opSet)
            -- this is lazy programming, not lazy evaluation!-)

                                    -- search for the best (1- or) 2-level tree
    alternatives _ [] bestML bestCTree bestIpParts bestOpParts =
      (bestCTree, bestIpParts, bestOpParts)  -- done

                             -- 2-level tree. NB simplest 0-lookahead algorithm
    alternatives n (sp:sps) bestML bestCTree bestIpParts bestOpParts =
      let theTree  = CTfork spMsgLen sp leaves    -- single fork
          -- spMsgLen = log (fromIntegral nSplts) -- crudest option
          spMsgLen = if nSplts < 8 then log (fromIntegral nSplts)  -- ad hoc
                     else nlPr wallaceIntModel n  -- bias to early splits
          leaves   = map CTleaf leafMdls          -- one leaf ...
          leafMdls = map estLeafMdl opParts       -- ... per part
          partNums = map (applySplitter sp) ipSet
          ipParts  = partition (aritySplitter sp) partNums ipSet
          opParts  = partition (aritySplitter sp) partNums opSet
          msgLen   = msg (functionModel2model theTree) (zip ipSet opSet)
          otherAlternatives = alternatives (n+1) sps
      in
         if msgLen < bestML                                   -- an improvement
         then otherAlternatives msgLen theTree   ipParts     opParts   -- new 1
         else otherAlternatives bestML bestCTree bestIpParts bestOpParts -- old

    splts  = splits ipSet  -- beware, splts very long => slow
    nSplts = length splts

   in case  alternatives 0 splts leafMsg leaf [ipSet] [opSet]  of
        ((CTfork msgLen pf leaves), ipParts, opParts) ->
          CTfork msgLen pf (zipWith search ipParts opParts) ;      -- subtrees?
        (t, _, _) -> t                                  -- the single leaf wins

 in search ipSet opSet
-- -------------------------------------------LA--http://www.allisons.org/ll/--
