-- See:
-- Lloyd ALLISON. Models for machine learning and data mining in
--   functional programming.       doi:10.1017/S0956796804005301
--   J. Functional Programming, 15(1), pp.15-32, Jan. 2005 (online 23 Jul 2004)
-- Lloyd ALLISON. Types and classes of machine learning and data mining.
--   ACSC2003, pp.207-215, February 2003.
-- Lloyd ALLISON. A programming paradigm for machine learning, with
--   a case study of Bayesian networks.  ACSC2006, January 2006.
--
-- Author: Lloyd ALLISON
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License (GPL) as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.  This program is distributed in the hope
-- that it will be useful, but without any warranty, without even the implied
-- warranty of merchantability or fitness for a particular purpose.  See the
-- GNU General Public License for more details.  You should have received a
-- copy of the GNU General Public License with this program; if not, write to:
-- Free Software Foundation, Inc., Boston, MA 02111, USA.

module IP.Classes (module IP.Classes) where                       -- export all

import IP.Utilities

--  SuperModel ---.--- Model dataSpace
--                |
--                |--- TimeSeries dataSpace
--                |
--                |--- FunctionModel inSpace opSpace
--                |
--                .--- ?other?

topline = "-- Inductive Programming 1.3 (IP_1.3), L.A., 8/2008 --"

type Probability   = Double
type MessageLength = Double

class (Show sMdl) => SuperModel sMdl where                        -- SuperModel
  prior    :: sMdl -> Probability    -- prior of blah-Model
  msg1     :: sMdl -> MessageLength  -- 1st part of any 2-part message
  msg1Base :: Double -> sMdl -> MessageLength
  mixture  :: (Mixture mx, SuperModel (mx sMdl)) =>
              mx sMdl -> sMdl        -- Mixture of blah_Model -> blah_Model

  prior sm = exp (-msg1 sm)
  msg1  sm = - log (prior sm)
  msg1Base b sm = (msg1 sm) / (log b)

  mixture mx = error "mixture not defined for SuperModel"
  -- NB. Any instance of SuperModel must define mixture and
  --     either prior or msg1 or both.
  -- In principle there might be a SuperModel not in class Show, but
  -- in practice easier to require them all to be, even if only trivially.



class Model mdl where                                                  -- Model
  pr   :: (mdl dataSpace) -> dataSpace -> Probability
  nlPr :: (mdl dataSpace) -> dataSpace -> MessageLength         -- neg log prob
  nlPrBase :: Double -> (mdl dataSpace) -> dataSpace -> MessageLength

  msg  :: SuperModel (mdl dataSpace) =>
          (mdl dataSpace) -> [dataSpace] -> MessageLength
  msgBase  :: SuperModel (mdl dataSpace) =>
              Double -> (mdl dataSpace) -> [dataSpace] -> MessageLength
  msg2 :: (mdl dataSpace) -> [dataSpace] -> MessageLength
  msg2Base :: Double -> (mdl dataSpace) -> [dataSpace] -> MessageLength

  pr   m datum = exp (- nlPr m datum) -- NB instances define pr or nlPr or both
  nlPr m datum = - log (pr m datum)
  nlPrBase b m datum = (nlPr m datum) / (log b)

  msg        m ds = (msg1 m) + (msg2 m ds)     -- complete 2-part message
  msgBase  b m ds = (msg  m ds) / (log b)
  msg2       m ds = foldl (+) 0 (map (nlPr m) ds)  -- part2  dataset|m  (list|m)
  msg2Base b m ds = (msg2 m ds) / (log b)
-- Some other properties (methods) that are possible, even desirable:
-- print (Show) yourself, generate sample data, ...


class TimeSeries tsm where                                        -- TimeSeries
-- given a TimeSeries, ts, of a dataSeries,  `predictors ts dataSeries' gives
-- a list of models for each position plus one (!) in the dataSeries,
-- each model conditional on the preceding data elements.
  predictors :: (tsm dataSpace) -> [dataSpace] -> [ModelType dataSpace]
  prs        :: (tsm dataSpace) -> [dataSpace] -> [Probability]
  nlPrs      :: (tsm dataSpace) -> [dataSpace] -> [MessageLength]
  nlPrsBase  :: Double -> (tsm dataSpace) -> [dataSpace] -> [MessageLength]
  -- any instance of TimeSeries must at least define  predictors
  -- and note that the output must be one longer than the dataSeries !
  -- context2model trialled for 2D DPA in 20040813, states seem better
  -- context2model :: (tsm dataSpace) -> [dataSpace]  -> ModelType dataSpace

  prs tsm dataSeries = zipWith pr (predictors tsm dataSeries) dataSeries

  nlPrs tsm dataSeries = zipWith nlPr (predictors tsm dataSeries) dataSeries

  nlPrsBase b tsm dataSeries =
    zipWith (nlPrBase b) (predictors tsm dataSeries) dataSeries

  -- context2model tsm = last.(predictors tsm).reverse  -- NB. most recent 1st
  -- context2model return a model/ prediction.  Ineff' for some TimeSeries.


class FunctionModel fm where                                   -- FunctionModel
  condModel :: (fm inSpace opSpace) -> inSpace -> ModelType opSpace
  condPr    :: (fm inSpace opSpace) -> inSpace -> opSpace -> Probability
  condNlPr  :: (fm inSpace opSpace) -> inSpace -> opSpace -> MessageLength
  condNlPrBase :: Double ->
               (fm inSpace opSpace) -> inSpace -> opSpace -> MessageLength
  -- any instance of FunctionModel must at least define  condModel

  condPr   fm i o = pr   (condModel fm i) o  -- cond prob, pr(o|i,fm)
  condNlPr fm i o = nlPr (condModel fm i) o
  condNlPrBase b fm i o = (condNlPr fm i o) / (log b)

-- -----------------------------------------------------------------------L.A--

class Mixture mx where                   -- for miscellaneous weighted mixtures
  mixer      :: (SuperModel t) => mx t -> ModelType Int
  components :: (SuperModel t) => mx t -> [t]

data (SuperModel elt) =>
  MixtureType elt = Mix (ModelType Int) [elt] -- weighted averages; MixtureType

instance Mixture MixtureType where
  mixer      (Mix m _ ) = m
  components (Mix _ es) = es

instance (SuperModel elt) => SuperModel (MixtureType elt) where
  msg1 (Mix m es) =
    foldl (+) (nlPr wallaceIntModel (length es - 1) + msg1 m) (map msg1 es)
  -- NB. This assumes that the receiver must be told the number of components.

                                                                   -- Printable
instance (SuperModel ct) => Show (MixtureType ct) where
  showsPrec p (Mix m es) s =
    "{Mix " ++ showsPrec p m (" : " ++ showsPrec p es ("}"++s))
  -- ct is in Show because SuperModels are in Show

-- Could have a MixtureModel, a sub-class of Mixture and of Model,
-- ditto MixtureTimeSeries, MixtureFunctionModel, but cannot(?)
-- have MixtureType as an instance of Model & TimeSeries & FunctionModel.
-- In any case,  mixture (Mix m cs)  is in the class of its components.

-- ----------------------------------------------------------------------------
                                                        -- conversion functions

model2timeSeries m =        -- :: Model of dataSpace -> TimeSeries of dataSpace
  TSM (msg1 m) (\context -> m) (\()->"TSM "++show m)   -- trivial but useful

                 -- :: Model of dataSpace -> FunctionModel of inSpace dataSpace
model2functionModel m = FM (msg1 m) (\ip -> m) (\()->"FM "++show m)


             -- Model of Int -> TimeSeries of dataSpace -> Model of [dataSpace]
             -- the 1st param is for "lengths"
timeSeries2model1 lenMdl tsm =
  MnlPr (msg1 tsm + msg1 lenMdl) (\dataSeries ->
            foldl (+) (nlPr lenMdl (length dataSeries)) (nlPrs tsm dataSeries))
            (\() -> "[" ++ (showsPrec 0 lenMdl (":" ++ (show tsm) ++ "]") ) )

timeSeries2model tsm =       -- TimeSeries of dataSpace -> Model of [dataSpace]
  timeSeries2model1 wallaceIntModel tsm


timeSeries2functionModel tsm =  -- TimeSeries of ds -> FunctionModel of [ds] ds
  FM (msg1 tsm) (\dataSeries -> last(predictors tsm dataSeries))
     (\()->"FM "++show tsm)


             -- FunctionModel of inSpace opSpace -> Model of (inSpace, opSpace)
functionModel2model fm = MnlPr (msg1 fm) (\(i, o) -> condNlPr fm i o)
                               (\() -> "(_,o):"++show fm)
                                -- i.e. pr(o|i,fm), i assumed common knowledge!

           -- FunctionModel of [dataSpace] dataSpace -> TimeSeries of dataSpace
functionModel2timeSeries fm =
  TSM (msg1 fm) (condModel fm) (\()->"TSM "++show fm)


  -- Model inSpace -> FunctionModel inSpace opSpace -> Model (inSpace, opSpace)
condition m fm =                          -- ? should the param order be v.v. ?
  let nlp (i,o) = (nlPr m i) + (nlPr (condModel fm i) o)
  in MnlPr (msg1 m + msg1 fm) nlp
           (\() -> "(" ++ showsPrec 0 fm ("|" ++ show m ++ ")"))

-- ----------------------------------------------------------------------------
                                                                   -- ModelType
data ModelType dataSpace =
  MPr   MessageLength (dataSpace -> Probability)   (() -> String) |
     -- msg len,      prob fn,                      description
  MnlPr MessageLength (dataSpace -> MessageLength) (() -> String)
     -- msg len,      neg log prob fn,              description

-- On wrappers :
-- With the existential (forall!) type extension, -fglasgow-exts, we can make
-- a ModelType "wrapper" option for other instances, m, of Model by adding
-- this to ModelType
  -- | forall m. (SuperModel (m dataSpace), Model m) => Mwrapper (m dataSpace)
-- this to the instance SuperModel
  -- msg1 (Mwrapper m) = msg1 m
-- this to instance Model
  -- nlPr (Mwrapper m) = nlPr m
-- this to instance Show
  -- show (Mwrapper m) = show m
-- etc., this will effectively get us mixtures of `m', as ModelType, for free.
-- Similarly FunctionModel, TimeSeries.

instance SuperModel (ModelType dataSpace)  where
  msg1 (MPr   mdlLen p _) = mdlLen
  msg1 (MnlPr mdlLen m _) = mdlLen
  mixture mx =
    let nlp datum =
          let (m0:ms) = components mx
              mxr     = mixer mx
              w n  []    total = total    -- done
              w n (m:ms) total =          -- next component, m
                w (n+1) ms (logPlus total ((nlPr mxr n)+(nlPr m datum)))
                                       --    component + datum|component
          in w 1 ms ((nlPr mxr 0) + (nlPr m0 datum))   -- do weighted average
    in MnlPr (msg1 mx) nlp (\()->show mx)

instance TimeSeries ModelType where   -- Model of ds is a trivial TimeSeries ds
  predictors m dataSeries = map (\_ -> m) ((error "") : dataSeries)
  -- Note that the ouput list is one longer than the input  dataSeries !

instance Model ModelType where
  nlPr (MPr   _ p _) datum = - log (p datum)
  nlPr (MnlPr _ n _) datum = n datum

instance Show (ModelType dataSpace)  where
  show (MPr   _ _ desc) = desc()
  show (MnlPr _ _ desc) = desc()

-- of course there can be other types that are instances of Model

-- ----------------------------------------------------------------------------
-- Logically, wallaceIntModel should be in module Models, but I get an
-- intermittent compile error if it is -- so there, or rather, here.

wallaceIntModel =                                -- Model of non-neg Int,  >= 0
-- A ``universal'' code, from the Wallace tree code (Wallace & Patrick 1993).
-- The number of full binary trees of n leaves is the (n-1)th Catalan number;
-- the nth catalan number is {[2n]C[n]}/(n+1)
-- C[0]=1, C[1]=1, C[2]=2, C[3]=5, C[4]=14, ...     -- n  code     cat' cum'
  let catalans =              -- NB. "cached"       -- 0  0        1    1
        let cats last n =                           -- 1  100      1    2
              let twoN = 2*n                        -- 2  10100    2    4
                  n1   = n+1                        -- 3  11000
                  nSq  = n*n                        -- 4  1010100  5    9
                  -- next = last * twoN * (twoN-1) `div` nSq  -- obvious
                  next = last * 2 * (twoN-1) `div` n          -- better
              in (next `div` n1) : (cats next n1)
        in 1 : (cats 1 1)
      cumulativeCatalans = scanl1 (+) catalans
      find n posn (e:es) = if n < e then posn else find n (posn+1) es
      nonNeg n = assert1 (0<=) n "wallaceIntModel, n is negative"
  in MnlPr 0 (\n -> ((find (nonNeg n) 0 cumulativeCatalans)*2+1)*log2)
           (\()->"wallaceIntModel")
-- Note that it is a non-parametric model.

-- ----------------------------------------------------------------------------
                                                              -- TimeSeriesType
data TimeSeriesType dataSpace =
  TSM MessageLength ([dataSpace] -> ModelType dataSpace) (() -> String)
               -- i.e. context   -> Model of next elt     description
               -- NB. context ordered last..first

instance SuperModel (TimeSeriesType dataSpace) where
  msg1 (TSM mdlLen m _) = mdlLen
  mixture mx =
    let f context = mixture(Mix (mixer mx)
                      (map (\(TSM _ ftsm _) -> ftsm context) (components mx)))
    in TSM (msg1 mx) f (\()->show mx)

instance TimeSeries TimeSeriesType where
  predictors (TSM mdlLen f _) dataSeries =
    let scan  []    context = [f context] -- NB. context ordered last..first
        scan (d:ds) context = (f context) : (scan ds (d:context))
    in scan dataSeries []
  -- This method of definition is efficient if f examines the last few datas.
  -- Note that the ouput list is one longer than the input  dataSeries !
  -- context2model (TSM _ f _) = f           -- trialled for 20040813

instance Show (TimeSeriesType dataSpace)  where
  show (TSM _ _ desc) = desc()

-- of course there can be other types that are instances of TimeSeries
-- ----------------------------------------------------------------------------
                                                           -- FunctionModelType
data FunctionModelType inSpace opSpace =
  FM MessageLength (inSpace -> ModelType opSpace) (() -> String)

instance SuperModel (FunctionModelType inSpace opSpace) where
  msg1 (FM mdlLen m _) = mdlLen
  mixture mx =
    let condM inp = mixture (Mix (mixer mx)
                                 (map (\f -> condModel f inp) (components mx)))
    in FM (msg1 mx) condM (\()->show mx)

instance FunctionModel FunctionModelType where
  condModel (FM mdlLen f _) = f

instance Show (FunctionModelType inSpace opSpace) where
  show (FM _ _ desc) = desc()

-- of course there can be other types that are instances of FunctionModel

-- -------------------------------------------LA--http://www.allisons.org/ll/--
