-- See:
-- Lloyd ALLISON. Models for machine learning and data mining in
--   functional programming.       doi:10.1017/S0956796804005301
--   J. Functional Programming, 15(1), pp.15-32, Jan. 2005 (online 23 Jul 2004)
-- Lloyd ALLISON. Types and classes of machine learning and data mining.
--   ACSC2003, pp.207-215, February 2003.
-- Lloyd ALLISON. A programming paradigm for machine learning, with
--   a case study of Bayesian networks.  ACSC2006, January 2006.
--
-- Author: Lloyd ALLISON
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License (GPL) as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.  This program is distributed in the hope
-- that it will be useful, but without any warranty, without even the implied
-- warranty of merchantability or fitness for a particular purpose.  See the
-- GNU General Public License for more details.  You should have received a
-- copy of the GNU General Public License with this program; if not, write to:
-- Free Software Foundation, Inc., Boston, MA 02111, USA.

module IP.FnModels (module IP.FnModels) where

import IP.Utilities
import IP.Classes
import IP.Models

-- Function Model (regression) - model of outputSpace given val in inputSpace.
-- NB. tuples of Enum Bounded types are made instances of Enum in Utilities.

                      -- (weighted) estimate a FunctionModel of ipSpace opSpace
estFiniteIpFunctionWeighted estOpModel ipSeries opSeries weights =
  let -- re weights: bug corrected 18/12/2003
      -- select outputs and weights corr' to a given input value, v.
      select v (ip:ips) (op:ops) (w:ws) selOp selW =
        if v == ip then select v ips ops ws (op:selOp) (w:selW)  -- pick
                   else select v ips ops ws     selOp     selW   -- drop
      select v _ _ _ selOp selW = (selOp, selW)
      mdls = map (\v -> uncurry estOpModel
                         (select v ipSeries opSeries weights [] []))
                 [lwbIp .. upbIp]
      part1 = foldl (+) 0 (map msg1 mdls)
      condMdl ip = mdls !! ((fromIp ip) - (fromIp lwbIp))

      fromIp ip = fromEnum (ip `asTypeOf` (ipSeries !! 0))
      lwbIp = minBound `asTypeOf` (ipSeries !! 0)
      upbIp = maxBound `asTypeOf` (ipSeries !! 0)
  in FM part1 condMdl (\()->"{finite_FM "++show mdls++"}")

estFiniteIpFunction estOpModelWeighted ipSeries opSeries =        -- unweighted
  estFiniteIpFunctionWeighted estOpModelWeighted ipSeries opSeries (repeat 1)


                                               -- outputSpace also Enum Bounded
estFiniteFunctionWeighted ipSeries opSeries weights =
 estFiniteIpFunctionWeighted estMultiStateWeighted ipSeries opSeries weights

estFiniteFunction ipSeries opSeries =                             -- unweighted
  estFiniteIpFunction estMultiStateWeighted ipSeries opSeries



         -- The next one will let us infer an order_k Markov model (TimeSeries)
         -- WARNING: It is assumed that |alphabet|**k is "small enough", else
         -- you had better implement ppm-like context trees or similar.
      -- estimate an order_k predictor FunctionModel of [dataSpace1] dataSpace2
estFiniteListFunction k inputs outputs =
  if k <= 0 then
    model2functionModel (estMultiState outputs)  -- order zero
  else  -- the order, k >= 1
    let select v []       _       selIps selOps = (selIps, selOps)     -- done
        select v _        []      selIps selOps = (selIps, selOps)     -- done
        select v ([]:ips) (d:ops) selIps selOps =    -- c empty
          select v ips ops selIps selOps             -- exclude
        select v (ip:ips)  (op:ops) selIps selOps =  -- c's matching [v,...]
          if v == head ip
          then select v ips ops ((tail ip):selIps) (op:selOps)  -- include
          else select v ips ops            selIps      selOps   -- exclude

        fms = map (\v -> uncurry (estFiniteListFunction (k-1))
                                 (select v inputs outputs [] []))
                  [lwbIp .. upbIp]  -- input values

        part1 = foldl (+) 0 (map msg1 fms)

        predictorFn [] =       -- k > |input|, i.e. input too short
          uniform lwbOp upbOp  -- at least it's simple!!!
        predictorFn (ip:ips) =
          condModel (fms !! ((fromIp ip) - (fromIp lwbIp))) ips

        egInput  = (inputs !! 0)!! 0
        upbIp    = maxBound `asTypeOf` egInput
        lwbIp    = minBound `asTypeOf` egInput
        fromIp ip = fromEnum( ip `asTypeOf` egInput )

        egOutput = outputs !! 0
        upbOp    = maxBound `asTypeOf` egOutput
        lwbOp    = minBound `asTypeOf` egOutput
        fromOp op = fromEnum( op `asTypeOf` egInput )

    in FM part1 predictorFn (\()->"{FiniteListFunction"++show fms++"}")

-- -------------------------------------------LA--http://www.allisons.org/ll/--

countFM countWeighted xs ys = countWeighted xs ys (repeat 1)
-- This operator could be used in other places than below.


-- This is the simplest univariate linear-regression as a case-study.
-- See lloyd's web pages, subsection ..../MML/Continuous/Linear1/
-- It is the intention to generalize linear models in the obvious ways.

linear1 eps a b s = linear1FM 0 eps a b s    -- a *fixed* linear1 FunctionModel

linear1FM m1 eps a b s =              -- y ~ N(a*x+b, s), simplest linear model
  -- eps is the data measurement accuracy, +/-(eps/2), of the dependent, y.
  FM m1 (\x -> normal eps (a*x+b) s)                    -- see normal in Models
     (\() -> "linear1(" ++ show a ++ "*x+" ++ show b ++
             "+N(0," ++ show s ++ ")(+-" ++ show eps ++ "))")

countLinear1 xs ys = countFM countLinear1Weighted xs ys

countLinear1Weighted xs ys ws =     -- gather the sufficient statistics
  let c n xSum ySum x2Sum xySum y2Sum (x:xs) (y:ys) (w:ws) =
        c (n+w) (x*w+xSum) (y*w+ySum)
                (x^2*w+x2Sum) (x*y*w+xySum) (y^2*w+y2Sum) xs ys ws
      c n xSum ySum x2Sum xySum y2Sum _ _ _ =
        (n, xSum, ySum, x2Sum, xySum, y2Sum)
  in c 0  0 0  0 0 0  xs ys ws

estLinear1 minB maxB minSigma maxSigma eps  xs ys =
  let (n, xSum, ySum, x2Sum, xySum, y2Sum) = countLinear1 xs ys
  in stats2linear minB maxB minSigma maxSigma eps n xSum ySum x2Sum xySum y2Sum

estLinear1Weighted minB maxB minSigma maxSigma eps  xs ys ws =
  let (n, xSum, ySum, x2Sum, xySum, y2Sum) = countLinear1Weighted xs ys ws
  in stats2linear minB maxB minSigma maxSigma eps n xSum ySum x2Sum xySum y2Sum

stats2linear minB maxB minSigma maxSigma eps  n xSum ySum x2Sum xySum y2Sum =
-- Mean y must lie in [minB, maxB].    NB The x values are "common knowledge".
-- Line passes through the CofG for {(xi,yi)}.     Beware "numerical" matters.
-- Cases with n < 3 are very dodgy indeed & most-sensible behaviour not clear.
  let xMean = xSum / n
      yMean = ySum / n
      a = if n <= 1 then 0.0                                -- 'a', slope
          else (xySum - xMean*ySum) / (x2Sum - xMean*xSum)
      b = if n < 1 then (minB + maxB) / 2.0   -- no data!
          else min maxB (max minB (yMean - a*xMean))        -- b, constant term
      -- Note, SUM{yi - a*xi - b}^2
      --     = SUM{yi^2 + (a*xi)^2 + b^2 - 2a*yi*xi - 2b*yi + 2a*b*xi}
      ySumSqDiff = y2Sum + a^2*x2Sum + n*b^2 - 2*(a*xySum + b*ySum - a*b*xSum)
      variance = ySumSqDiff / (n-1)
      sigma = min maxSigma (max (max eps minSigma) (if n <= 1
              then maxSigma         -- ??? no information about sigma
              -- then sqrt(minSigma * maxSigma)  -- GM, also see 'normal' Model
              else if n == 2        -- is 'a' zero or is sigma zero?!
              then sqrt(y2Sum - yMean*ySum)      -- over n-1 == 1
              else sqrt variance ))
      nlPriorA = 1/(pi*(1 + a^2))  --i.e. uniform on the angular(!) slope
      nlPriorB = log(maxB - minB)  --i.e. uniform on the range
      nlPriorSigma  = log(log maxSigma - log minSigma)+log sigma  --i.e 1/sigma
      nlPrior   = nlPriorA + nlPriorB + nlPriorSigma
      konstant  = (1 + log(latticeLimit))        --what is lattice K3 ???
      logFisher = log 2 + 2*log n + log(x2Sum - xMean*xSum) - 6*log sigma
      msg1 = if n <= 1 then 0.0   -- "silly" cases
             else nlPrior + logFisher/2 + konstant
  in linear1FM msg1 eps a b sigma
-- -------------------------------------------LA--http://www.allisons.org/ll/--
