-- See:
-- Lloyd ALLISON. Models for machine learning and data mining in
--   functional programming.       doi:10.1017/S0956796804005301
--   J. Functional Programming, 15(1), pp.15-32, Jan. 2005 (online 23 Jul 2004)
-- Lloyd ALLISON. Types and classes of machine learning and data mining.
--   ACSC2003, pp.207-215, February 2003.
-- Lloyd ALLISON. A programming paradigm for machine learning, with
--   a case study of Bayesian networks.  ACSC2006, January 2006.
--
-- Author: Lloyd ALLISON
-- This program is free software; you can redistribute it and/or modify it
-- under the terms of the GNU General Public License (GPL) as published by
-- the Free Software Foundation; either version 2 of the License, or (at
-- your option) any later version.  This program is distributed in the hope
-- that it will be useful, but without any warranty, without even the implied
-- warranty of merchantability or fitness for a particular purpose.  See the
-- GNU General Public License for more details.  You should have received a
-- copy of the GNU General Public License with this program; if not, write to:
-- Free Software Foundation, Inc., Boston, MA 02111, USA.

module IP.Models (module IP.Models) where

import IP.Sci        --for logFactorial and integrate
import IP.Utilities
import IP.Classes

bivariate (m1, m2) =         -- (Model of d1, Model of d2) -> Model of (d1, d2)
  let nlp (d1, d2) = (nlPr m1 d1) + (nlPr m2 d2)  -- d1, d2 assumed independent
  in MnlPr ((msg1 m1) + (msg1 m2)) nlp
           (\() -> "(" ++ showsPrec 0 m1 ("," ++ show m2 ++ ")") )

estBivariate (est1, est2) dataSet =
  let (ds1, ds2) = unzip dataSet
  in bivariate (est1 ds1, est2 ds2)

estBivariateWeighted (est1, est2) dataSet weights =    -- for mixture modelling
  let (ds1, ds2) = unzip dataSet
  in bivariate (est1 ds1 weights, est2 ds2 weights)

estCondition estM estFM xys =                            -- bivariate where y|x
  let { (xs,ys)=unzip xys;  mx = estM xs;  fm = estFM xs ys }
  in condition mx fm               -- tested with FunctionModels in next module

estConditionWeighted estM estFM xys ws =        -- weighted bivariate where y|x
  let { (xs,ys)=unzip xys;  mx = estM xs ws;  fm = estFM xs ys ws }
  in condition mx fm
-- ----------------------------------------------------------------------------

-- 2003, I moved     wallaceIntModel, -- Model of non-neg Int,  >= 0
-- to module Classes for "pragmatic" reasons.


quadraticIntModel =        -- a simple parameterless Model of non-neg Int, >= 0
  let nlPr n = if n < 0 then error "n < 0"
               else log (fromIntegral ((n+1)*(n+2)))  -- n >= 0
                    -- i.e.  pr n = 1/((n+1)*(n+2))
  in MnlPr 0 nlPr (\() -> "P(n)=1/((n+1)*(n+2), n>=0")
  -- NB. Sum[n=0..] 1/((n+1)*(n+2)) = Sum[n=0..]{ 1/(n+1) - 1/(n+2) }
  --     = 1 - 1/2 + 1/2 - 1/3 + 1/3 ... = 1,  i.e. it is a prob' dist'n.


fifty50 = uniform 0 1                        -- 50:50 on {0,1}


uniform lo hi =                              -- Model of Enum Bounded dataSpace
  let (lwb, upb) = if (fromEnum hi) >= (fromEnum lo)
                   then (lo, hi) else (hi, lo)  -- ?switch?
  in MPr 0 (\_ -> 1 / fromIntegral((fromEnum upb) - (fromEnum lwb) + 1))
         (\() -> "uniform[" ++ show lo ++ ".." ++ show hi ++ "]")


probs2model ps =                           --  :: [Probability] -> Model of Int
  MPr 0 (\n -> ps !! n ) (\() -> "mState"++show ps)  -- NB. MPr 0 ...


freqs2model fs =     -- :: frequencies, [Num] -> Model of Int (i.e. MultiState)
  -- Taking a short-cut here, see Wallace and Boulton (1968), and
  -- Wallace and Freeman (1987) for more information.
  let total = foldl (+) 0 fs
      -- noModel estimates cost of data using the "uninformative" method
      noModel  []    ans = ans
      noModel (f:fs) ans = noModel fs (ans - (logFactorial f))   -- NB base e
      -- MML:  part1 + part2 = noModel... + delta, so...
      part1 = (noModel fs (logFactorial (total+1))) + delta - part2
      part2 = - (foldl (+) 0 ((zipWith (*) fs (map log ps))))
            -- part2 = msg Len of data given the MML model
      delta = nParams * ((log(pi/6))+1) / 2 -- about 0.176 nits per parameter
      nParams = fromIntegral(length fs - 1)
      ps = normalise (map ((+) 0.5) fs)  -- probabilities, NB +0.5  MML est'r
      p n = ps !! n                      -- the prob' function
  in MPr (if part1 > 0 then part1 else 0) p (\()->"mState"++show ps)


       -- list of frequencies from a series of (weights and) Enum, Bounded type
countWeighted dataSeries weights =
  let c  []     _     counters = counters
      c (d:ds) (w:ws) counters = c ds ws (inc counters (fromEnum d) w)
      inc  []    n _ = error ("count, in inc, ran off end of list "++(show n))
      inc (c:cs) 0 w = (c+w) : cs            -- counter, c += w
      inc (c:cs) n w = c : (inc cs (n-1) w)  -- w is a weight
      zeroes 0 = []
      zeroes n = 0 : (zeroes (n-1))  -- e.g. zeroes 2 = [0, 0]
      upb = maxBound `asTypeOf` (dataSeries !! 0)
      lwb = minBound `asTypeOf` (dataSeries !! 0)
      typeSize = (fromEnum upb) - (fromEnum lwb) + 1
  in c dataSeries weights (zeroes typeSize)

count dataSeries =   -- list of frequencies from a series of Enum, Bounded type
  countWeighted dataSeries (repeat 1)  -- totalAssignment



                      -- (Bounded t, Enum t) => t -> Model of Int -> Model of t
modelInt2model egValue intModel =       -- example value gives the Model's type
  let fromE x = fromEnum( x `asTypeOf` egValue )
      toInt x = (fromE x) - (fromE minBound)
      p datum = pr intModel (toInt datum)
  in MPr (msg1 intModel) p (\()->show intModel)



estMultiStateWeighted dataSet weights =    -- weights and dataSet -> Model of t
  modelInt2model (dataSet !! 0) (freqs2model (countWeighted dataSet weights))

estMultiState dataSet =                    -- :: [Enum Bounded t] -> Model of t
  modelInt2model (dataSet !! 0) (freqs2model (count dataSet))



coinByPrH prH = modelInt2model H (probs2model [prH, 1-prH])

-- -------------------------------------------------9/2002--7/2007--L.Allison--

geometric mu = geometricModel 0 mu     -- i.e. a *fixed* geometric distribution
  -- NB. datum >= 0, and parameter is the mean.

geometricModel msg1 mu =
  MnlPr msg1 (geometricNlPr mu) (\() -> "G0(" ++ show mu ++ ")")

geometricNlPr mu d =                                       -- NB. d >= 0, not 1
  let d' = fromIntegral d
  in (d'+1) * log(mu+1) - d' * log mu              -- NB. parameter is the mean

estGeometric a ds = stats2geometric a (countInts ds)

estGeometricWeighted a ds ws = stats2geometric a (countIntsWeighted ds ws)

stats2geometric a (total, n) =
  let h mu = exp(-mu / a) / a                  --prior, has mean = a
      logF = log n - log mu - log(mu+1)        --log Fisher
      msg1 = max 0 (log a + mu/a + logF/2 - (log 12 - 1)/2)
      mu = (1 - n - 1/a
           + sqrt(n*n + 1/(a*a) + 1 + 2*n/a - 2*n + 4*total/a)) * a/2     --MML
  in geometricModel msg1 mu

-- -------------------------------------------------9/2002--7/2007--L.Allison--

poisson alpha = poissonModel 0 alpha     -- i.e. a *fixed* Poisson distribution
  -- Note, datum >= 0, and parameter is the mean

poissonModel msg1 alpha =
  MnlPr msg1 (poissonNlPr alpha) (\() -> "P(" ++ show alpha ++ ")")

poissonNlPr alpha d =                -- Note, Poisson's mean = variance = alpha
  let d' = fromIntegral d
  in  alpha - d'*(log alpha) + logFactorial d'                        -- d >= 0

estPoisson a ds = stats2Poisson a (countInts ds)

estPoissonWeighted a ds ws = stats2Poisson a (countIntsWeighted ds ws)

stats2Poisson a (total, n) =
  let h alpha = exp(-alpha / a) / a            --prior, has mean = a
      logF    = log n - log alpha              --log Fisher
      msg1    = max 0 (- log(h alpha) + logF/2 - (log 12 - 1)/2)
      -- msg1    = max 0 (- log(h alpha) - 0.5*log(12 * alpha / n))
      alpha   = (total + 0.5) / (n + 1/a)      --MML estimate
  in poissonModel msg1 alpha

countInts ds = countIntsWeighted ds (repeat 1)                 --get suff.stats

countIntsWeighted ds ws =
  let c [] _ stats = stats
      c (d:ds) (w:ws) (total, nData) =
        c ds ws (total + (fromIntegral d)*w, nData + w)
  in c ds ws (0, 0)

-- ---------------------------------------------------------7/2007--L.Allison--

intNormal m s = intNormalModel 0 m s  --i.e. a fixed integer (quantised) Normal

intNormalModel m1 m s =
  MnlPr m1 (intNormalNlPr m s) (\() -> "intN("++show m++","++show s++")")

intNormalPr m s x =                                    -- a bit rough and ready
  let x' = fromIntegral x
  in integrate (x'-0.5) (x'+0.5) 10 (\x' -> exp(-normalNlPrDensity m s x'))

intNormalNlPr m s x =                                                  -- ditto
  let x' = fromIntegral x
  in if 3*s >= abs(x'-m) then -log(intNormalPr m s x) else normalNlPr 1 m s x'

estIntNormal minMean maxMean minSigma maxSigma xs =
  estIntNormalWeighted minMean maxMean minSigma maxSigma xs (repeat 1)

estIntNormalWeighted minMean maxMean minSigma maxSigma xs ws =
  let xs' = map fromIntegral xs
      (n, xSum, xSumSqDiff) = countCtsWeighted xs' ws
      (msg1, mean, sigma) =
        stats2normalParams minMean maxMean minSigma maxSigma 1          --eps=1
                                   n xSum xSumSqDiff
  in intNormalModel msg1 mean sigma

-- -------------------------------------------LA--http://www.allisons.org/ll/--

normal eps m s = normalModel 0 eps m s                 -- i.e. a *fixed* Normal

normalModel m1 eps m s =                         -- i.e. Gaussian a.k.a. Normal
  -- eps is the data measurement accuracy,  +/-(eps/2).
  if s < eps then  -- need better integration, much earlier really!!!
       error "normal: std-dev too small v. data measurement accuracy"
  else MnlPr m1 (normalNlPr eps m s)                          -- Model of Float
       (\() -> "N(" ++ show m ++ "," ++ show s ++ ")(+-" ++ show eps ++ "/2)")

normalNlPr    eps m s x = normalNlPrDensity m s x - log eps    -- neg log prob'

normalNlPrDensity m s x =                                      -- neg log dens'
  let constPart = (log(2 * pi)) / 2 + (log s)
  in constPart + (((x-m)/s)**2)/2


countCtsWeighted dataSeries weights =         -- "count" a continuous attribute
  let n  = sum (take (length dataSeries) weights)
      s  = sum (zipWith (*) dataSeries weights)           -- I know, multi-pass
      sq = sum (zipWith (\x -> \w -> w * sqr(x-m)) dataSeries weights)
      m  = s / n
  in (n, s, sq)                              -- (number, sum, sum squared diff)

countCts dataSeries = countCtsWeighted dataSeries (repeat 1)    -- weights == 1


estGaussianWeighted = estNormalWeighted  -- synonyms

estGaussian = estNormal                  -- synonym

estNormalWeighted minMean maxMean minSigma maxSigma eps xs ws =
  let (n, s, sq) = countCtsWeighted xs ws
  in stats2normal minMean maxMean minSigma maxSigma eps  n s sq

estNormal minMean maxMean minSigma maxSigma eps xs =
  let (n, s, sq) = countCts xs
  in stats2normal minMean maxMean minSigma maxSigma eps  n s sq


stats2normal minMean maxMean minSigma maxSigma eps  n xSum xSumSqDiff =
  let (msg1, mean, sigma) = stats2normalParams minMean maxMean
                              minSigma maxSigma eps  n xSum xSumSqDiff
  in normalModel msg1 eps mean sigma

stats2normalParams minMean maxMean minSigma maxSigma eps  n xSum xSumSqDiff =
-- Dealing with the normal (Gaussian) distribution is a good deal trickier
-- than sometimes realised.  The programmer cannot have any prior idea of the
-- ranges of the mean and standard deviation, but the guy with the data does!
-- So we require information about the ranges to be passed in.
-- This version of an estimator uses a uniform prior for the mean over
-- the given range, and 1/sigma over the given range.
--   (In some applications it is not unknown to stretch correctness and
--    to use priors based on the population (data) statistics.)
-- Priors, mean:  uniform over [minMean..maxMean]
--         sigma: 1/sigma over [minSigma..maxSigma]
-- Data is measured to +/-eps/2
--
-- f(x|m,s) = (1/(sqrt(2 pi) s)) exp(-sqr(x-m)/(2 sqr(s)))        pr(x) density
-- L = -ln(f( ))                                         neg log likelihood(xi)
--   = N.{ln(2 pi)/2 + ln(s)} + {SUMi (xi-m)^2}/(2 s^2)
-- 1st derivatives:
--   d L/d m = {N.m - SUMi xi}/s^2         (hence max LH est', m ~ sum/N)
--   d L/d s = -N/s - {SUMi (xi-m)^2}/s^3  (hence max LH est', s ~.../N)
-- 2nd derivatives:
--   d2 L/d m d s = zero in expectation
--   d2 L/d m2    = N/s^2
--   d2 L/d s2    = N/s^2 + 3.{SUM (xi-m)^2}/s^4   = - 2.N/s^2 in expectation
-- Fisher Info = 2.N^2 / s^4      (e.g. see Ch6 s6.2 CSW's book-draft 6/2003)
-- NB. This code is deliberately straightforward, not optimised.
--
 let xMean = xSum / n
     xVariance = xSumSqDiff / (n-1)                               -- NB n-1

     mean  = if n < 1
             then (minMean + maxMean) / 2.0        -- no data!
             else min maxMean (max minMean xMean)  -- force within limits
     sigma = if n <= 1
             then maxSigma  -- ??? hard to know, no information about sigma
             -- then exp((log minSigma + log maxSigma) / 2) -- geometric mean
             else min maxSigma (max (max eps minSigma) (sqrt xVariance))

-- 1. It does not make sense to infer a standard deviation less
--    than something-like(!) the data measurement accuracy.
-- 2. A better(!) method of integrating the prior and the
--    likelihood should be used if n is small.
-- 3. Sometimes (small) data-sets are "odd", with just the one value
--    repeated in which case should you really be using the normal?
-- It is always these special cases that cause the most difficulty!

     nlPriorMean   = log(maxMean - minMean)               -- i.e. uniform prior
       -- normalised, 1/(maxMean - minMean)
-- NB. INTEGRAL[a..b] 1/x = log b - log a
     nlPriorSigma  = log(log maxSigma - log minSigma)+log sigma  -- i.e 1/sigma
       -- normalised to 1/(sigma*(log(maxSigma - log minSigma))
     nlPrior       = nlPriorMean + nlPriorSigma
-- konstant: the *2, for 2 parameters, cancels with /2 in the general msg form.
     konstant      = (1 + log(latticeK2))
     halfLogFisher = (log 2)/2 + log n - 2 * log sigma
     logFisher     = 2 * halfLogFisher
-- msg1, 1st part of message, e.g., Farr 1999, or,
-- e.g., see lloyd's MML pages ..../MML/Notes/Fisher/
     -- msg1 = if n <= 0 then 0.0  -- a "silly" case, 2003
     msg1 = if n < 1 then 0.0  -- a "silly" case, 11/2005
            else nlPrior + halfLogFisher + konstant

 in (msg1, mean, sigma)
-- -------------------------------------------LA--http://www.allisons.org/ll/--
