--Prototype source code for numerical scientific computing in Haskell, ghc.
--Here: 1. some standard functions,
--      2. array manipulation,
--      3. function optimisation,
--      4. constraints.
--Also see Sci.hs, and T.hs.
--Compile tests thus: ghc --make Sci  (has run under ghc 6.4.1, Feb 2008).

--Terminology:
--  "vector" means a 1-D array,
--  "matrix" means a 2-D array, and
--  "array"  is the most general term, arbitrary # of dimensions.

module IP.Sci (module IP.Sci) where

import Array  --a standard library.
--For Haskell's 'Array' library, see...
--  http://haskell.org/ghc/docs/latest/html/libraries/array/Data-Array.html

-- -----------------------------------2/2008--LA--http://www.allisons.org/ll/--

integrate = simpson  --until anything better

simpson lo hi n f =  --Simpson's rule integrate f over [lo,hi] with n intervals
  let n'     = fromIntegral n
      range  = if hi >= lo then hi-lo else 0.0
      coeffs = cycle [4.0, 2.0]  -- == [4,2,4,2,4,...]
      fs  = [ f( (lo*(n'-i) + hi*i)/n' ) | i <- map fromIntegral [1..n-1] ]
      ans = (f lo + f hi + sum (zipWith (*) coeffs fs)) * range / (3*n')
  in if n > 1 then if even n then ans else simpson lo hi (n+1) f
     else error("simpson ... n=" ++ show n)


stirling       n =                                  -- Stirling's approximation
  let ans = (n + 0.5)*(log n) - n + (log(2*pi)) / 2          -- ~logBase e (n!)
  in max ans 0                                             -- also see logGamma

stirlingBase b n = (stirling n) / (log b)   -- and to base b


logFactorial x = logGamma(x+1) -- x>0, much much better than Stirling's approx.

logGamma x =         --logGamma(x|x>0), logGamma x = (x-1)!, logGamma(x+1) = x!
  let p = [0.99999999999980993,  676.5203681218851,    -1259.1392167224028,
           771.32342877765313,  -176.61502916214059,    12.507343278686905,
          -0.13857109526572012,  9.9843695780195716e-6, 1.5056327351493116e-7]
      g = 7
      s = (head p) + sum(zipWith (/) (tail p) [x+i | i <- [0..]])
      t = x + g - 0.5
  in if x <= 0 then error("gamma " ++ show x)                      -- x <= 0
     else if x < 0.5 then log pi - log(sin(pi*x)) - logGamma(1-x)  --reflection
     else 0.5*log(2*pi) + (x-0.5)*log t - t + log s                -- x >= 0.5

psi0 = digamma                                     --1st derivative of logGamma
digamma = let dx = sqrt(delta 1) in \x ->
  if x <= 0 then error("digamma " ++ show x)
  else if x < 0.5 then  - pi / tan(pi*x) + digamma(1-x)   -- !
  else diffEps (x*dx) logGamma x                   -- x >= 0.5
--NB. x > 0;  digamma aka psi or psi0;  digamma = d/dx logGamma,
--e.g., digamma 1.0 = (the constant) -gamma = -0.577,215,664,9... .

--or... digamma = let dx = sqrt(delta 1) in \x -> diffPosEps (x*dx) logGamma x
--But this version is rough and ready.

psi1 = trigamma                                    --2nd derivative of logGamma
trigamma = let dx = (delta 1)**0.25; pi2 = pi*pi in \x ->
  if x <= 0 then error("trigamma " ++ show x)
  else if x < 0.5 then pi2/(sin(pi*x))**2 - trigamma(1-x)  -- !
  else diff2Eps (x*dx) logGamma x                   -- x >= 0.5
--NB. x > 0;  trigamma aka psi1;  trigamma = d2/dx2 logGamma = d/dx digamma,
--e.g., psi1 1.0 = pi**2/6 = 1.6449341.

--or...trigamma = let dx = (delta 1)**0.25 in \x -> diffPosEps (x*dx) digamma x
--But this version is rough and ready.
--trigamma x = diffPosEps (x/1000) digamma x  --the bog simplest alternative.

delta one =          --Find x s.t. one-x != one but one-x'/8 = one, apparently.
  let f iter x =  --NB. delta is polymorphic; type can be Float or Double etc..
        (if iter > 20 || one-x' == one then x else f (iter+1) x')
        where x' = x/8
  in f 1 (one/8)                           --find limits of numerical accuracy.
-- ----------------------------------------------------------------------------

dot v1 v2 =                                 --vector dot (inner) product, v1.v2
  let (b1, b2) = (bounds v1, bounds v2)
  in if b1 == b2 then sum [v1!i * v2!i | i <- range b1]
     else error("dot product " ++ show b1 ++ " !~ " ++ show b2)
  --Takes what seems the most general option with regards to the bounds.
  --NB. v1 and v2 could even be multi-dimensional if that were useful.

normaliseArray v =                                  --return w~v s.t. |w|=w.w=1
  let (s, w) = absAndNormalise v
  in if s > 0 then w else error "normaliseArray: magnitude of vector is <= 0"

absAndNormalise v =           --return (|v|,w) s.t. w~v and |w|=1, if possible!
  let s = sqrt(sum [vi*vi | i<-range(bounds v), let vi=v!i]) --equiv (v.v)**0.5
  in (s, if s > 0 then fmap (\x -> x/s) v else v)

row m r =                              --select the vector row r of 2D matrix m
  let ((_,lo), (_,hi)) = bounds m  --requires m to be 2D
  in ixmap (lo,hi) (\c -> (r,c)) m --map result!c onto m!(r,c) for the given r

col m c =                           --select the vector column c of 2D matrix m
  let ((lo,_), (hi,_)) = bounds m  --requires m to be 2D
  in ixmap (lo,hi) (\r -> (r,c)) m --map result!r onto m!(r,c) for the given c

mTranspose m =                                          --transpose a 2D matrix
  let ((lo1,lo2),(hi1,hi2)) = bounds m  --requires m to be 2D
  in ixmap ((lo2,lo1),(hi2,hi1)) (\(r,c) -> (c,r)) m
  --NB. The Haskell standard library 'List' contains a 'transpose' on [[t]].


-- c.f. Matlab tends to treat (convert) most things as a 2D vector and
-- to treat vectors as row (1xn) or column (nx1) 2D matrices as necessary.
-- The next two functions are unnecessary if we do not take that path.
vec2row v =                       --turn a 1D vector into a 2D 1x|v| row-matrix
  let (bLo, bHi) = bounds v
  in ixmap ((0,bLo),(0,bHi)) (\(r,c) -> if r==0 then c else error "vec2row") v

vec2col v =                    --turn a 1D vector into a 2D |v|x1 column-matrix
  let (bLo, bHi) = bounds v
  in ixmap ((bLo,0),(bHi,0)) (\(r,c) -> if c==0 then r else error "vec2col") v


arrayZipWith f a1 a2 =   --apply f to corr. pairs of elements of arrays a1 & a2
  let b1 = bounds a1
      b2 = bounds a2
  in if b1 == b2 then listArray b1 [f (a1!i) (a2!i) | i <- range b1]
     else error ("arrayZipWith " ++ show b1 ++ " !~ " ++ show b2)
  --a pity the standard class Functor only has the one "method" fmap.

--ghc, running as h98, requires the following type sig. for arrayPlus etc. ...
arrayPlus :: (Ix i, Show i, Num t) => (Array i t) -> (Array i t) -> (Array i t)
arrayPlus  = arrayZipWith (+)                --add corr. elements of two arrays

arrayMinus:: (Ix i, Show i, Num t) => (Array i t) -> (Array i t) -> (Array i t)
arrayMinus = arrayZipWith (-)           --subtract corr. elements of two arrays

saTimes s arr =               --multiply every element of array arr by scalar s
  fmap ((*) s) arr

vmTimes v m =                           --multiply "row" vector and matrix, v*m
  let vB = bounds v
      mB = bounds m
      ((mLo1,mLo2), (mHi1,mHi2)) = mB
      mB2 = (mLo2, mHi2)
  in if vB == (mLo1, mHi1) then       --conformable
       listArray mB2 [dot v (m `col` c) | c <- range mB2]
     else error("vmTimes " ++ show vB ++ " !~ " ++ show mB)

mvTimes m v =                        --multiply matrix and "column" vector, m*v
  let vB = bounds v
      mB = bounds m
      ((mLo1,mLo2), (mHi1,mHi2)) = mB
      mB1 = (mLo1, mHi1)
  in if vB == (mLo2, mHi2) then       --conformable
       listArray mB1 [dot (m `row` r) v | r <- range mB1]
     else error("mvTimes " ++ show mB ++ " !~ " ++ show vB)

mmTimes mA mB =                                  --matrix multiplication, mA*mB
  let (bA, bB) = (bounds mA, bounds mB)     --bounds
      ((aLo,aHi), (bLo,bHi)) = (bA, bB)
      (r1, r2, r3) = ( (fst aLo, fst aHi),  --index ranges
                       (snd aLo, snd aHi),
                       (snd bLo, snd bHi) )
  in if r2 == (fst bLo, fst bHi) then       --conformable
       array ((fst r1, fst r3), (snd r1, snd r3))
         [((i,k), (mA `row` i) `dot` (mB `col` k)) | i<-range r1, k<-range r3]
     else error("mmTimes " ++ show bA ++ " !~ " ++ show bB)


det :: (Num t) => Array (Integer, Integer) t -> t                 --Num for (*)
det = naiveDeterminant                                --determinant of matrix m
  --could do better.

naiveDeterminant :: (Num t) => Array (Integer, Integer) t -> t    --Num for (*)
naiveDeterminant m =                                  --determinant of matrix m
  let b = bounds m           --naive alg. is slow and of questionable stability
      ((lo1,lo2), (hi1,hi2)) = b
  in if hi1 >= lo1 && hi1-lo1 == hi2-lo2 then  --square
       if hi1 == lo1 then m!(lo1,lo2)          --1x1 matrix
       else foldr1 (-) [m!(lo1,c)*det(dropRowCol m lo1 c) | c<-range(lo2,hi2)]
     else error("naiveDeterminant " ++ show b ++ " matrix")

dropRowCol m r c =                      --drop row r, and column c, of matrix m
  let ((lo1,lo2),(hi1,hi2)) = bounds m
  in ixmap ((lo1,lo2), (hi1-1,hi2-1))
           (\(r',c') -> (if r' < r then r' else r'+1,
                         if c' < c then c' else c'+1)) m
-- ----------------------------------------------------------------------------

diffEps    eps f x = (f(x+eps) - f(x-eps))/(2*eps)  --slope, finite differences
diffPosEps eps f x = (f(x+eps) - f x)/eps    --finite differences, +ve side
diffNegEps eps = diffPosEps (-eps)           --finite differences, -ve side!
                                                   --NB. x and (f x) are scalar

diff2Eps    eps f x = (f(x+eps)   - 2*f x + f(x-eps))/(eps*eps)  -- 2nd deriv
diff2PosEps eps f x = (f(x+2*eps) - 2*f(x+eps) + f x)/(eps*eps)  -- +ve side
diff2NegEps eps = diff2PosEps (-eps)                             -- -ve side!


diffAdaptive = diffAdaptiveDfEps diffEps 0.001       --adaptive differentiation
  --Can also use the Pos and Neg diffs with diffAdaptive.
  --Nothing magic about the default eps.

diffAdaptiveEps :: (Ord r, Fractional r) => r -> (r->r) -> r -> r
diffAdaptiveEps = diffAdaptiveDfEps diffEps     --most often use symmetric diff
  --NB. Ord r for the comparison, and Fractional r for (/).

diffAdaptiveDfEps d eps f x =     --NB. x is a scalar, adaptive differentiation
--Use when unsure what eps to use; initial eps must be reasonably small to be
--in the monotonic region (note, safe if f, or f', is a constant function);
--d can be the symmetric, Pos or Neg diff...Eps.
--See Stepleman & Winarsky, Math. of Comput., 33(148), pp.1257-1264, 1979.
  let df dx      = d dx f x               --approx f' = d/dx f by finite diffs
      approx dx  = df dx : approx (dx/4)  --try decreasing epsilons
      d1:(d2:ds) = approx eps    --tend monotonically to f' until m/c acc limit
      monotone   = compare d1 d2 --LT|EQ|GT, note that EQ is handled specially
      adapt d12 d2 (d3:ds) =
        let d23 = abs (d2-d3)    --monotonically decreases until m/c acc limit
        in if d23 <= d12 && compare d2 d3 == monotone then adapt d23 d3 ds
           else d2               --d3 affected by m/c (in)accuracy, so take d2
  in if monotone==EQ then d1 else adapt (abs (d1-d2)) d2 ds


grad :: (Ix i, Fractional r) => (Array i r -> r) -> (Array i r) -> (Array i r)
grad = gradEps 0.001   --use default eps; returns the k-dim. gradient of f at x
  --NB. Fractional r for (/).

gradEps eps f x =           --grad returns the k-dimensional gradient of f at x
  let bs = bounds x         --NB. f's parameter, x, is a continuous vector
  in listArray bs [diffEps eps (\xi -> f (x//[(i,xi)])) (x!i) | i<-range bs]
  --Now, it is handy to have some convenience fns to convert
  --f(x,y), and similar, into g(Array...), see arrayFrom* and arrayTo*

gradAdaptive :: (Ix i, Ord r, Fractional r) =>
  (Array i r -> r) -> (Array i r) -> (Array i r)
gradAdaptive = gradAdaptiveEps 0.001      --just use a default eps
  --NB. type signature corr. to diffAdaptive.

gradAdaptiveEps eps f x =                 --grad using adaptive Differentiation
  let bs = bounds x       --here, the same starting eps for each dimension of x
  in listArray bs
       [diffAdaptiveEps eps (\xi -> f (x//[(i,xi)])) (x!i) | i<-range bs]


arrayFromT2 (x,y) = listArray (0,1) [x, y]               -- (t,t) <-> Array 2xt
arrayToT2 a = let (lo,hi) = bounds a
              in if hi-lo == 1 then (a!lo, a!hi)
                 else error "needed an array of 2 elements"

arrayFromT3 (x,y,z) = listArray (0,2) [x, y, z]        -- (t,t,t) <-> Array 3xt
arrayToT3 a = let (lo,hi) = bounds a
              in if hi-lo == 2 then (a!lo, a!(lo+1), a!hi)
                 else error "needed an array of 3 elements"

arrayFromT4 (w,x,y,z) = listArray (0,3) [w, x, y, z] -- (t,t,t,t) <-> Array 4Xt
arrayToT4 a = let (lo,hi)= bounds a
              in if hi-lo == 3 then (a!lo, a!(lo+1), a!(lo+2), a!hi)
                               else error "needed an array of 4 elements"

class Arrayable t where      --a type, t, s.t. t can be converted to/from Array
  toVector   :: t e -> Array Int e
  fromVector :: Array Int e -> t e

instance Arrayable [] where                                --list, [] <-> Array
  toVector xs = listArray (0, (length xs)-1) xs  --standard bounds (0,...)
  fromVector = elems

-- Tuples, sadly: Kind error: `(,)' is not applied to enough type arguments
--   In the instance declaration for `Arrayable (,)'.
--   (The `kind' of (,) differs from [].)
-- instance Arrayable (,) where  --Array 2xt <-> (t,t)
  -- toVector (x,y) = listArray (0,1) [x, y]
  -- fromVector a = let (lo,hi) = bounds a
                -- in if hi-lo == 2 then (a!lo, a!hi)
                   -- else error "array !~ (,)"
-- and using the type-synonym  type T2 e = (e,e)  is no help either.


hessian :: (Ix i, Fractional r) =>
  (Array i r -> r) -> (Array i r) -> Array (i,i) r
hessian = hessianEps 0.001       --matrix of 2nd derivatives of function f at x
hessianEps eps f x =             --matrix of 2nd derivatives of function f at x
  let bs = bounds x  --NB. f's parameter, x, is a vector of k floats
      (lo, hi) = bs
      eps = 0.001    --by finite differences
      d2 = eps*eps
      h = listArray ((lo,lo),(hi,hi))
        [ hij | i<-range bs, j<-range bs,
          let hij = if i > j then h!(j,i)       --symmetric!
                else if i == j then              --diagonal
                  let xp = x//[(i, x!i+eps)]
                      xm = x//[(i, x!i-eps)]
                  in (f xp - 2*f x + f xm)/d2
                else                             -- j > i
                  let xpp = x//[(i, x!i+eps), (j, x!j+eps)]
                      xpm = x//[(i, x!i+eps), (j, x!j-eps)]
                      xmp = x//[(i, x!i-eps), (j, x!j+eps)]
                      xmm = x//[(i, x!i-eps), (j, x!j-eps)]
                  in (f xpp + f xmm - f xpm - f xmp)/(4*d2) ]
  in h
  --NB. Fractional r for (/).


argmin1 :: (Ix i, Show i, Ord t, Floating t) =>  --for ghc (h98)
  (Array i t -> Array i t -> (Bool, Array i t)) -> (Array i t -> t)
  -> (Array i t -> Array i t) -> Array i t -> Array i t
argmin1 = argmin1step 1         --minimise a function subject to constraint(s).
--The 1 stands for 1st order; defaults will not suit every use!
--Default step0 = 1.

argmin1step step0 constraint f f' x0 =
--Minimise f by gradient descent subject to constraint(s);
--the 1 stands for 1st order; defaults will not suit every use!
--You can rewrite stopping, e.g., to give an error if iter>100, if you like.
  (\(_,x,_) -> x) (gradientDescent step0 stopping constraint f f' x0)
    where stopping iter step = step < 0.0001 || iter > 100

gradientDescent step0 stopping constraint f f' x0 =             --simplest alg.
--Seek x to minimise (f x) subject to constraint;  x0 is initial given est.
--Pre: x0 passes the constraint.  Returns (|lastStep|, x, f x).
--Can use (grad f) or similar for f' if you do not have a better fn for it,
--but beware of f' blowing-up if constraint not satisfied.
  let fmin iter step x0 fx0 =
        let (sz, g)    = absAndNormalise(f' x0)  --g is direction to move x
            (unco, x1) = constraint x0 (x0 `arrayMinus` (step `saTimes` g))
            fx1 = f x1
        in if sz==0 || stopping iter step then (step, x0, fx0)        --done
           else if fx1 > fx0 then fmin (iter+1) (step/2) x0 fx0       --higher
           else fmin (iter+1) (if unco then step else step/2) x1 fx1  --LE
  in fmin 1 step0 x0 (f x0)


listConstraint []     _  x1 = (True, x1)                       --listConstraint
listConstraint (c:cs) x0 x1 =
--Check that x1 satisfies each one of a list of constraints and,
--if not, modify it so that it does (may use x0).
--Pre: (c x0) is true, and modification of x1 to satisfy c will not
--invalidate any previous constraint, and will not be invalidated by
--any later constraint in cs; this is the case for axisConstraints,
--e.g., [x!0 > 0, x!1 < 10],
  let (uncon, x2) = c x0 x1
      (uncon',x3) = listConstraint cs x0 x2
  in (uncon && uncon', x3)

nullConstraint x0 x1 = (True, x1)   --trivial, always satisfied, without change

axisConstraint i cnd bnd x0 x1 =        --specify axis i's upper or lower bound
--Boundary constraint solved by 'projection constraint strategy';
--if x1 fails, make a satisfactory x1' as close to x1 as possible.
--cnd::[Ordering]; must not return the bnd itself unless EQ in cnd.
--Most commonly cnd is one of [LT], [LT,EQ], [GT,EQ], or [GT],
--but [EQ], [LT,GT] (i.e. not=), [LT,EQ,GT] (trivial) are also possible.
  if (compare (x1!i) bnd) `elem` cnd then (True, x1)  --x1 ok as is
  else (False, x1')    --must modify x1 for constraint satisfaction
    where x1' = x1//[(i, if EQ `elem` cnd then bnd else (9*bnd+x0!i)/10)]  -- !

axisConstraint2 i cnd1 lb cnd2 ub =       --specify both lower and upper bounds
  listConstraint [axisConstraint i cnd1 lb, axisConstraint i cnd2 ub]
  --a convenience function

-- -------------------------------LA--for-use-with-Inductive-Programming-1.2+--
-- -------------------------------http://dx.doi.org/10.1017/S0956796804005301--
