--Student's t-distribution.
--Also see Sci.hs and T.hs.
--Compile tests thus: ghc --make T    (has run under ghc 6.4.1, Feb 2008).

module IP.T where
--Should import T "qualified" as function names are general to all models.
--T.model, T.est and T.est_wtd are the main intended exports, & not the_works!

import IP.Sci           --locally defined
import IP.Utilities(sqr, latticeK3)
import IP.Classes  (ModelType(..))

import Array         --a standard library
import List(sortBy)  --for x0, to start the search to est. the t-distribution.

-- -----------------------------------2/2008--LA--http://www.allisons.org/ll/--

                                                  --Student's t-distribution...
model m1 eps (params@(_,s,_)) =
--Forms a fixed Student's t-distribution (Model) with the given params;
--m1 is the model's message length;
--eps is the width of the data measurement (uncertainty) region.
  if s < eps then                  -- must come back to this
    error "t-distribution: s < data measurement uncertainty"
  else MnlPr m1 (nlPr eps params)
             (\() -> "t" ++ show params ++ "(+/-" ++ show eps ++ "/2)")


nlPr eps params x = nlPrDensity params x - log eps   --negative log probability
--Note this is a true probability;  it assumes that the
--probability density varies little across [x-eps/2,x+eps/2].

nlPrDensity (m,s,v) x =                      --negative log probability density
-- f(x|m,s,v), for m (mu) location, s (sigma) spread, v (nu) shape parameter.
  let v1  = v+1
      s2  = s*s
      vs2 = v*s2
      constPart =  0.5*log(pi*v) + logGamma(v/2) + log s - logGamma(v1/2)
  in constPart + (v1/2) * log(1 + sqr(x-m) / vs2)


-- mu sigma nu ,-)                                      negative log likelihood
negLogLH (m, s, v) dataSet = negLogLH_wtd(m, s, v) dataSet (repeat 1)

negLogLH_wtd (m, s, v) dataSet wts =                  --negative log likelihood
--Negative log likelihood (density) for Student's t-distribution
--for *weighted* data.
--v=1 gives a Cauchy;  v=30 ~ Normal distribution (Gaussian).
--The variance is v*s**2/(v-2), if v>2, otherwise undefined.
  let c = 0.5*log pi + logGamma(v/2) -logGamma((v+1)/2) -(v/2)*log v - v*log s
      vs2 = v*s*s
      (n, _, _, sumLogXm2) = stats_wtd (m, s, v) dataSet wts
  in n*c + ((v+1)/2)*sumLogXm2

gradNegLogLH (m, s, v) dataSet = gradNegLogLH_wtd (m, s, v) dataSet (repeat 1)
                                           --negative log likelihood's gradient
gradNegLogLH_wtd (m, s, v) dataSet wts =
  let (n, sum1Xm2, sumXmXm2, sumLogXm2) = stats_wtd (m, s, v) dataSet wts
  in listArray (0,2)
       [ -(v+1)*sumXmXm2,                 -- d/d m,
         -n*v/s + v*(v+1)*s*sum1Xm2,      -- d/d s,  and last is d/d v
         n*(0.5*digamma(v/2) - 0.5*digamma((v+1)/2) - 0.5 - 0.5*log v - log s)
         + 0.5*sumLogXm2 + ((v+1)*s*s/2)*sum1Xm2]


stats_wtd (m, s, v) dataSet wts =
--NB. These are not "sufficient statistics", and they depend on the parameters.
  let n  = sum(take (length dataSet) wts)  -- !
      vs2 = v*s*s
      sum1Xm2   = sum(zipWith (\x -> \w ->         w / (vs2 + (x-m)**2))
                      dataSet wts)
      sumXmXm2  = sum(zipWith (\x -> \w -> w * (x-m) / (vs2 + (x-m)**2))
                      dataSet wts)
      sumLogXm2 = sum(zipWith (\x -> \w ->     w * log (vs2 + (x-m)**2))
                      dataSet wts)
  in (n, sum1Xm2, sumXmXm2, sumLogXm2)


fisher n (m, s, v) =     --Student's t-distribution's Fisher information matrix
--A 3x3 matrix.  Say you held v constant, that would leave the top-left 2x2.
  let vp1   = v+1                            --    m     s     v
      vp3   = v+3                            -- |d2m2    0     0   | --m
      vp3s2 = vp3*s*s                        -- |  0    d2s2 d2dsdv| --s
      d2dm2 = (vp1 / vp3s2)*n                -- |  0  d2dsdv  d2dv2| --v
      d2ds2 = (2*v / vp3s2)*n
      d2dv2 = (0.25*(psi1(v/2) - psi1(vp1/2)) - (v+5)/(2*v*vp1*vp3))*n
      d2dvds= -2/(vp1*vp3*s)*n
  in listArray ((0,0),(2,2)) [d2dm2,0,0, 0,d2ds2,d2dvds, 0,d2dvds,d2dv2]
  --Commonly one wants logF, as below.

logF n (m, s, v) =    --log determinant of the (full) Fisher information matrix
  let vp1 = v+1                                 --    m     s     v
      vp3 = v+3                                 -- |d2m2    0     0   | --m
      -- vp3s2 = vp3*s*s                        -- |  0    d2s2 d2dsdv| --s
      -- d2dm2 = vp1 / vp3s2                    -- |  0  d2dsdv  d2dv2| --v
      -- d2ds2 = 2*v / vp3s2
      -- d2dv2 = 0.25*(psi1(v/2) - psi1(vp1/2)) - (v+5)/(2*v*vp1*vp3)
      -- d2dvds= -2/(vp1*vp3*s)
  -- in 3*log n + log d2dm2 + log(d2ds2 * d2dv2 - d2dvds * d2dvds)   --equiv...
  -- in log(determinant(fisher n (m,s,v)))                           --equiv...
  in 3*log n - 4*log s
     + log( v*vp1/(2*vp3*vp3)*(psi1(v/2)-psi1(vp1/2)) - 1/(vp1*vp3) )

gradLogF (m, s, v) =                                                 --gradLogF
--Note, data is iid, so n is not involved.
  let vp1 = v+1
      vp3 = v+3
      f v = log( v*vp1/(2*vp3*vp3)*(psi1(v/2) - psi1(vp1/2)) - 1/(vp1*vp3) )
  in listArray (0,2) [0, -4.0/s, diffPosEps 0.001 f v]


start dataSet = start_wtd dataSet (repeat 1)                            --start
                  --Get a reasonable starting point, (m0,s0,v0), for the search
start_wtd dataSet wts =
--The median is a robust estimator for m,  s is trickier,  and
--everyone says to start v (nu, shape) at 6.
  let mean   = (sum (zipWith (*) dataSet wts)) / n'
      varnc  = (sum (zipWith (\x-> \w-> w*(x-mean)**2) dataSet wts)) / n'
      stdDev = (delta 1) `max` (sqrt varnc)  --must be greater than 0
      median = f 0 (sortBy(\(x,_) -> \(x',_) ->compare x x')(zip dataSet wts))
        where f runTot ((x,w):xws) = if runTot+w >= n'/2 then x
                                     else f (runTot+w) xws
      n' = sum (take n wts)  -- !
      n  = length dataSet
  in if      n' <= 0 then (0, 1, 6)
     else if n' <= 1 then (median, 1, 6)
     else (median, stdDev, 6)  --a "reasonable" starting pt for optimisation.
  --Actually the stdDev+/- would also be a reasonable initial step size.


    --Maximum Likelihood (MLH) parameter estimator for Student's t-distribution
est_MLH_params dataSet = est_MLH_params_wtd dataSet (repeat 1)   -- all wts=1.0

est_MLH_params_wtd dataSet wts =
  let ta params = negLogLH_wtd (arrayToT3 params) dataSet wts
      tg params = gradNegLogLH_wtd (arrayToT3 params) dataSet wts
      startPoint = arrayFromT3 (start_wtd dataSet wts)
      step0 = startPoint!1
  in arrayToT3 (argmin1step step0 constraint0 ta tg startPoint)


   --c.f. the MLH one, the MML parameter estimator for Student's t-distribution
est_MML_params h constraints dataSet =
  est_MML_params_wtd h constraints dataSet (repeat 1)    --wts=1.0

est_MML_params_wtd h constraints dataSet wts =
  snd( the_works 0.001 (arrayFromT3(0,1,6)) h constraints dataSet wts )


                         --Minimum Message Length MML Model-estimator for the t
est minMean maxMean minSigma maxSigma eps dataset =
  est_wtd minMean maxMean minSigma maxSigma eps dataset (repeat 1)

                                                        --and for weighted data
est_wtd minMean maxMean minSigma maxSigma eps xs ws =
-- eps is width of the data measurement region
  let h  = prior       minMean maxMean minSigma maxSigma
      cn = constraints minMean maxMean minSigma maxSigma
      (_,sp) = cn (arrayFromT3((minMean+maxMean)/2, (minSigma+maxSigma)/2, 6))
	          (arrayFromT3 (start_wtd xs ws))
      ((m1, _), params) = the_works eps sp h cn xs ws
  in model m1 eps params


the_works eps startPoint h constraints dataSet wts =               -- the_works
--NB. h is the prior distribution on (m,s,v),
--eps is the data measurement accuracy (also limits est. precision).
  let log_k3 = latticeK3     --lattice constant, for 3 parameters.
      n = sum(take (length dataSet) wts)  -- !
      m1m2 msv =  --calculate message length
        let m1 = 0 `max` (-log(h msv) + 0.5*logF n msv + 1.5*log_k3 + 1.5)
            m2 = negLogLH_wtd msv dataSet wts
            --NB. ignoring data accuracy, eps, here because it "passes through"
        in (m1, m2)            --i.e., (model, (data|model)).
      msgLen = (uncurry (+)) . m1m2   --total message length.
      gradMsgLen msv =
        (gradEps (eps/4) (fst.m1m2.arrayToT3) (arrayFromT3 msv)) `arrayPlus`
        (gradNegLogLH_wtd msv dataSet wts)  --dominant (& time consuming) term
      step0      = 1 `max` (startPoint!1)  --at least 1 for v (nu)
      est = arrayToT3 (argmin1step step0 constraints
                         (msgLen.arrayToT3) (gradMsgLen.arrayToT3) startPoint)
  in ((m1m2 est), est)


constraint0 =                                --a basic constraint on the params
  listConstraint [
    axisConstraint  1 [GT] 0,                    -- s > 0, sigma
    axisConstraint2 2 [GT,EQ] 1.0 [LT,EQ] 30]    -- 30 >= v >= 1, nu

constraints min_m max_m min_s max_s =        --gives a constraint on the params
  listConstraint [
    axisConstraint2 0 [GT,EQ] min_m [LT,EQ] max_m,   -- 0: m, mu
    axisConstraint2 1 [GT,EQ] min_s [LT,EQ] max_s,   -- 1: s, sigma
    axisConstraint2 2 [GT,EQ] 1.0   [LT,EQ] 30 ]     -- 2: 30 >= v >= 1, nu

prior min_m max_m min_s max_s (m,s,v) =                --gives a standard prior
-- (it corresponds to the default usage of the normal (Gaussian) distribution).
  let pm = 1 / (max_m - min_m)                 -- uniform on [min_m,max_m]
      ps = 1 / (s * (log max_s - log min_s))   -- 1/s     on [min_s,max_s]
        --Note, Integral[a,b] 1/s ds = log b - log a.
      pv = 1 / 29                              -- uniform on [1,30]
  in pm * ps * pv
--You can use a different prior and/or constraint if you want to.


--NB. For data x, prior h, likelihood f, Fisher F, and 3 params (k3=0.078543),
--the total message length =
--  -log(h theta) - log f(x|theta) + (1/2)(log F) + (3/2)(1+log k3)
--See lloyd's pages ..../MML/Notes/Fisher/ on minimum message length (MML).
-- --------------LA--for-use-with-http://dx.doi.org/10.1017/S0956796804005301--
